/****************************************************  Common Functions  *****************************************************

	Red Rapids
	797 North Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following functions are common to all SigFPGA hardware and channel types.

	Rev History:  11-11-2015:	Modified the ClockInit() function to exclude the Model 276 from extended searches.
				  08-21-2015:	Modified the ClockInit() function to correct problems with the expanded search using the AD9512.
				  05-08-2015:	Changed bit assignments for the LED and LoadGPIOSettings functions.
				  08-21-2014:	Inital release.

	Copyright 2015, Red Rapids, Inc.

****************************************************  Common Functions  *****************************************************/
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "adapter.h"
#include "sigstream.h"
#include "dma_functions.h"
#include "common_functions.h"
#include "channel_functions.h"
#include "device_functions.h"

// Close the Red Rapids device on the PCI bus.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long BusClose(s_Adapter *p_HW)
{
	// Close the assigned device.
	return(Adapter_Close(p_HW));
}

// Open the Red Rapids device on the PCI bus.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long BusOpen(s_Adapter *p_HW)
{
	// Set all of the global interrupt mask register values in the hardware handle.
	p_HW->GlobalIntMask.BAR = BAR0;
	p_HW->GlobalIntMask.Offset = GLOBAL_INT_MASK;
	p_HW->GlobalIntMask.Disable = 0x0ull;
	// Open the assigned device.
	return(Adapter_Open(p_HW));
}

// Report the number of receiver (0) or transmitter (1) channels available in the device.
// Returns the number of channels.
unsigned long ChannelCount(s_Adapter *p_HW, size_t RxTxSelect)
{
	unsigned long long Value = 0;

	Read(p_HW, BAR2, IDENTIFICATION, &Value);

	return((unsigned long)((Value >> (44 + 4*RxTxSelect)) & 0x000000000000000Full));
}

// Synchronize the converter clocks.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long ClockInit(s_Adapter *p_HW)
{
	unsigned long Status = 0;
	unsigned long long Value = 0;
	unsigned long long Busy = 0;
	unsigned long FineAdjust = 0;
	size_t watchdog = 0;

	// Wait for clock lock to find a solution across the available search window.
	do
	{
		// Reset the watchdog counter.
		watchdog = 0;

		// Command the clock lock process to start.
		Status += Write(p_HW, BAR0, CLOCK_INITIALIZE, 0x0ull);

		// Wait for the clock lock process to begin.
		while (!Busy && (watchdog < 100))
		{
			// Check the clock lock busy status.
			Status += Read(p_HW, BAR0, CLOCK_INITIALIZE, &Busy);
			// Increment the watchdog counter.
			watchdog++;
		}

		// Wait for clock lock to find a solution across a 2.5 ns search window.
		do
		{
			// Save and clear the contents of the error status register.
			Status += Write(p_HW, BAR0, ERROR_STATUS, 0x0ull);
			// Read the saved copy of the error status register to determine if all converter clocks are locked.
			Status += Read(p_HW, BAR0, ERROR_STATUS, &Value);
			// The search failed if the process stops without a solution.
			Status += Read(p_HW, BAR0, CLOCK_INITIALIZE, &Busy);
			// Increment the watchdog counter.
			watchdog++;
		}
		while ((Value & 0x0000000000000F00ull) && Busy && (watchdog < 100));

		// Phase shift the converter clock and start a new search if a solution was not found.
		// Note: Not all products support this feature.
		if ((ModelNumber(p_HW) != 0x276) && (Value & 0x0000000000000F00ull))
		{
			// Disable the delay bypass in the clock distribution chip.
			Status += AD9512Write(p_HW, DEFAULT_CLK_DIST, 0x0034, 0x00);
			// Load the fine phase adjust with an integer from 0 to 31.
			// Note: Start from zero since the delay block includes an inherent 600 ps offset.
			Status += AD9512Write(p_HW, DEFAULT_CLK_DIST, 0x0036, FineAdjust << 1);
			// Perform a register update to commit the new phase delay values.
			Status += AD9512Write(p_HW, DEFAULT_CLK_DIST, 0x005A, 0x01);
			// Increment the delay approximately 300 ps for another possible search.
			FineAdjust++;
		}
	}
	while ((ModelNumber(p_HW) != 0x276) && (Value & 0x0000000000000F00ull) && (FineAdjust != 32) && (watchdog < 100));

	// Return an error if no solution was found.
	if (Value & 0x0000000000000F00ull) Status++;

	return(Status);
}

// Report the current status of all clocks.
// Returns the following status codes:
// 0: External sample clock active, digital clocks are locked.
// 1: External sample clock active, digital clocks are not locked.
// 2: Sample clock locked to internal reference, digital clocks are locked.
// 3: Sample clock locked to internal reference, digital clocks are not locked.
// 4: Sample clock locked to external reference, digital clocks are locked.
// 5: Sample clock locked to external reference, digital clocks are not locked.
// >5: Sample clock is not locked.
unsigned long ClockStatus(s_Adapter *p_HW, unsigned long ChipSelect)
{
	unsigned long long Value = 0;
	unsigned long Status = 0;
	
	if (!((unsigned long long)FXL6408Read(p_HW, ChipSelect, 0xF) & 0x0000000000000002ull))
		// External sample clock active.
		Status = 0;
	else
		if (!((unsigned long long)FXL6408Read(p_HW, ChipSelect, 0xF) & 0x0000000000000001ull))
			//Sample clock not locked.
			Status = 6;
		else
			if ((unsigned long long)FXL6408Read(p_HW, ChipSelect, 0xF) & 0x0000000000000004ull)
				// Sample clock locked to internal reference.
				Status = 2;
			else
				// Sample clock locked to external reference.
				Status = 4;

	// Save and clear the contents of the error status register.
	Write(p_HW, BAR0, ERROR_STATUS, 0x0ull);
	// Read the saved copy of the error status register.
	Read(p_HW, BAR0, ERROR_STATUS, &Value);

	if (Value & 0x0000000000000030ull)
		// An error was reported on the digital clock status.
		Status++;

	return(Status);
}

// Set the error interrupt mask bits to enable (1) or disable (0) interrupt generation for each error type.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long ErrorMask(s_Adapter *p_HW, s_ErrorStatus ErrorIntMask)
{
	return(	Write(p_HW, BAR0, ERROR_INT_MASK,
				(unsigned long long)ErrorIntMask.Bar[0] |
				(unsigned long long)ErrorIntMask.Bar[1] << 1 |
				(unsigned long long)ErrorIntMask.Bar[2] << 2 |
				(unsigned long long)ErrorIntMask.Bar[3] << 3 |
				(unsigned long long)ErrorIntMask.Clock[0] << 4 |
				(unsigned long long)ErrorIntMask.Clock[1] << 5 |
				(unsigned long long)ErrorIntMask.Clock[2] << 6));
}

// Report the firmware revision date (MMDDYYYY).
// Returns the firmware revision date in hex format, but it is not a hex value (e.g. 04/27/2012 returns 0x04272012).
unsigned long FirmwareRevision(s_Adapter *p_HW)
{
	unsigned long long Value = 0;
	
	Read(p_HW, BAR2, IDENTIFICATION, &Value);
	
	return((unsigned long)(Value & 0x00000000FFFFFFFFull));
}

// Report the error status and clear the values.
// Returns the total number of errors detected.
unsigned long GetErrorStatus(s_Adapter *p_HW, s_ErrorStatus *p_ErrorStatus)
{
	unsigned long long Value = 0;
	unsigned long Status = 0;
	size_t cnt = 0;
	
	// Save and clear the contents of the error status register.
	Write(p_HW, BAR0, ERROR_STATUS, 0x0ull);
	// Read the saved copy of the error status register.
	Read(p_HW, BAR0, ERROR_STATUS, &Value);

	// Strip the individual BAR error indicator bits.
	for (cnt = 0; cnt < 4; cnt++)
		if ((p_ErrorStatus->Bar[cnt] = (Value >> cnt) & 0x0000000000000001ull)) Status++;

	// Strip the individual clock error indicator bits.
	for (cnt = 0; cnt < 8; cnt++)
		if ((p_ErrorStatus->Clock[cnt] = (Value >> (cnt + 4)) & 0x0000000000000001ull)) Status++;

	// Return the total error count.
	return(Status);
}

// Report the state of channel and error status flags.
// Returns the total number of flags that require service.
unsigned long GetGlobalStatus(s_Adapter *p_HW, s_GlobalStatus *p_GlobalStatus)
{
	unsigned long long Value = 0;
	unsigned long Status = 0;
	size_t cnt = 0;
	
	// Read the global status.
	Read(p_HW, BAR0, GLOBAL_STATUS, &Value);

	// Strip the individual channel status flags.
	for (cnt = 0; cnt < 4; cnt++)
		if ((p_GlobalStatus->Channel[cnt] = (Value >> cnt * 4) & 0x0000000000000001ull)) Status++;

	// Strip the error status flag.
	if ((p_GlobalStatus->Error = (Value >> 32) & 0x0000000000000001ull)) Status++;

	// Return the total active flag count.
	return(Status);
}

// Retrieve bus voltage [0] and current [1] on all power supplies and calculate power dissipation [3].
// Returns total power dissipation.
double GetPower(s_Adapter *p_HW, s_Power *p_Power)
{
	// Read the positive bus voltage [0] and shunt voltage [1] on all three supplies and calculate power [3].
	p_Power->Supply1[0] = (INA3221Read(p_HW, VOLT_MONITOR, 0x02) >> 3) * 8.0e-3;
	p_Power->Supply1[1] = ((INA3221Read(p_HW, VOLT_MONITOR, 0x01) >> 3) * 40.0e-6) / .01;
	p_Power->Supply1[2] = p_Power->Supply1[0] * p_Power->Supply1[1];
	p_Power->Supply2[0] = (INA3221Read(p_HW, VOLT_MONITOR, 0x04) >> 3) * 8.0e-3;
	p_Power->Supply2[1] = ((INA3221Read(p_HW, VOLT_MONITOR, 0x03) >> 3) * 40.0e-6) / .01;
	p_Power->Supply2[2] = p_Power->Supply2[0] * p_Power->Supply2[1];
	p_Power->Supply3[0] = (INA3221Read(p_HW, VOLT_MONITOR, 0x06) >> 3) * 8.0e-3;
	p_Power->Supply3[1] = ((INA3221Read(p_HW, VOLT_MONITOR, 0x05) >> 3) * 40.0e-6) / .01;
	p_Power->Supply3[2] = p_Power->Supply3[0] * p_Power->Supply3[1];

	return(p_Power->Supply1[2] + p_Power->Supply2[2] + p_Power->Supply3[2]);
}

// Retrieve temperature from the device.
// Returns the total number of errors detected.
unsigned long GetTemperature(s_Adapter *p_HW, s_Temperature *p_Temperature)
{
	unsigned long long one = 0x1, max = 0xFFFFFFFFFFFFFFFFull;
	unsigned long Register[6] = {0,0,0,0,0,0};
	unsigned long Status = 0;
	size_t cnt = 0;
	
	// Configure all of the LTC2991 ports (V1-V8) for temperature measurement.
	Status += LTC2991Write(p_HW, TEMP_MONITOR, 0x06, 0x22);
	Status += LTC2991Write(p_HW, TEMP_MONITOR, 0x07, 0x22);
	// Perform a conversion on all of the ports and internal sensor.
	Status += LTC2991Write(p_HW, TEMP_MONITOR, 0x01, 0xF8);
	// Wait for the conversion process to complete by monitoring the busy bit.
	do
		MicrosecondTimer(50000);
	while(LTC2991Read(p_HW, TEMP_MONITOR, 0x01) & 0x00000004);
	// Retrieve the temperature values.
	Register[0] = (LTC2991Read(p_HW, TEMP_MONITOR, 0x0A) << 8) | LTC2991Read(p_HW, TEMP_MONITOR, 0x0B);
	Register[1] = (LTC2991Read(p_HW, TEMP_MONITOR, 0x0E) << 8) | LTC2991Read(p_HW, TEMP_MONITOR, 0x0F);
	Register[2] = (LTC2991Read(p_HW, TEMP_MONITOR, 0x12) << 8) | LTC2991Read(p_HW, TEMP_MONITOR, 0x13);
	Register[3] = (LTC2991Read(p_HW, TEMP_MONITOR, 0x16) << 8) | LTC2991Read(p_HW, TEMP_MONITOR, 0x17);
	Register[4] = (LTC2991Read(p_HW, TEMP_MONITOR, 0x1A) << 8) | LTC2991Read(p_HW, TEMP_MONITOR, 0x1B);
	Register[5] = (LTC2991Read(p_HW, TEMP_MONITOR, 0x1C) << 8) | LTC2991Read(p_HW, TEMP_MONITOR, 0x1D);
	// Check the data valid bit on each register.
	for (cnt = 0; cnt < 6; cnt++)
		if (Register[cnt] & 0x00008000)
		{
			Register[cnt] &= 0x000001FFF;
			// Sign extend negative numbers.
			if (Register[cnt] & (one << 12)) Register[cnt] |= (max << 13);
			// Apply any necessary diode ideality correction and convert each measurement to degrees Celsius.
			switch (cnt)
			{
				case 0: {p_Temperature->Sensor[0] = ((1.004/1.010)*((Register[0]/16)+273.14))-273.15; break;}
				case 1: {p_Temperature->Sensor[1] = Register[1]/16; break;}
				case 2: {p_Temperature->Sensor[2] = Register[2]/16; break;}
				case 3: {p_Temperature->Sensor[3] = Register[3]/16; break;}
				case 4: {p_Temperature->Sensor[4] = Register[4]/16; break;}
				case 5: {p_Temperature->Vcc = 2.5+(Register[5]*315.18e-6); break;}
				default: {Status++; break;}
			}
		}
		else
			Status++;

	return(Status);
}

// Read the current time of day (TOD) seconds value.
// Note:  The fractional seconds value is not available.
// Returns the TOD seconds value.
unsigned long GetTODSeconds(s_Adapter *p_HW)
{
	unsigned long long Value = 0;
	
	Read(p_HW, BAR2, TOD_SETTINGS, &Value);

	return((unsigned long)(Value & 0x00000000FFFFFFFFull));
}

// Issue a global synchronization strobe that is synchronous to all channels.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long GlobalSync(s_Adapter *p_HW)
{
	return(Write(p_HW, BAR2, GLOBAL_SYNC, 0x0ull));
}

// Set the global interrupt mask to enable (1) or disable (0) interrupt generation from the device.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long InterruptMask(s_Adapter *p_HW, unsigned long GlobalIntMask)
{
	return(Write(p_HW, BAR0, GLOBAL_INT_MASK, (unsigned long long)GlobalIntMask));
}

// Select the sources that will illuminate LED-A and LED-B.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long LED(s_Adapter *p_HW, s_LEDSource LEDSource)
{
	return(	Write(p_HW, BAR2, LED_SOURCE_MASK,
		((unsigned long long)LEDSource.LEDA[0] & 0x1ull) |
		((unsigned long long)LEDSource.LEDA[1] & 0x1ull) << 1 |
		((unsigned long long)LEDSource.LEDA[2] & 0x1ull) << 2 |
		((unsigned long long)LEDSource.LEDA[3] & 0x1ull) << 3 |
		((unsigned long long)LEDSource.LEDA[7] & 0x1ull) << 4 |
		((unsigned long long)LEDSource.LEDA[0] & 0x2ull) << 4 |
		((unsigned long long)LEDSource.LEDA[1] & 0x2ull) << 5 |
		((unsigned long long)LEDSource.LEDA[2] & 0x2ull) << 6 |
		((unsigned long long)LEDSource.LEDA[3] & 0x2ull) << 7 |
		((unsigned long long)LEDSource.LEDA[7] & 0x2ull) << 8 |
		((unsigned long long)LEDSource.LEDA[8] & 0x1ull) << 10 |
		((unsigned long long)LEDSource.LEDA[9] & 0x1ull) << 11 |
		((unsigned long long)LEDSource.LEDA[9] & 0x2ull) << 11 |
		(unsigned long long)LEDSource.LEDB[0] << 14 |
		(unsigned long long)LEDSource.LEDB[1] << 15 |
		(unsigned long long)LEDSource.LEDB[2] << 16 |
		(unsigned long long)LEDSource.LEDB[3] << 17));
}

// Configure the GPIO ports and set the logic state of any that are programmed as outputs.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long LoadGPIOSettings(s_Adapter *p_HW, s_GPIOSettings GPIO)
{
	unsigned long Status = 0;
	unsigned long Value = 0;
	unsigned long Direction = 0;

	// Tristate all of the voltage translators before changing the signal direction to avoid contention.
	Status += Write(p_HW, BAR2, GPIO_SETTINGS, 0x1ull);

	// Set the GPIO voltage translator direction.
	Direction =
		(unsigned long long)GPIO.Direction[5] << 18 |
		(unsigned long long)GPIO.Direction[4] << 19 |
		(unsigned long long)GPIO.Direction[3] << 20 |
		(unsigned long long)GPIO.Direction[2] << 21 |
		(unsigned long long)GPIO.Direction[1] << 22 |
		(unsigned long long)GPIO.Direction[0] << 23;

	// Load the translator direction settings into the port expander.
	Status += Serial(p_HW, GPIO_USER_XLATOR, RISING_EDGE, 0x0, 0, Direction, 24, &Value);

	// Load the GPIO configuration.
	Status += Write(p_HW, BAR2, GPIO_SETTINGS,
		(unsigned long long)GPIO.Direction[0] |
		(unsigned long long)GPIO.Direction[1] << 1 |
		(unsigned long long)GPIO.Direction[2] << 2 |
		(unsigned long long)GPIO.Direction[3] << 3 |
		(unsigned long long)GPIO.Direction[4] << 4 |
		(unsigned long long)GPIO.Direction[5] << 5 |
		(unsigned long long)GPIO.OutputValue[0] << 6 |
		(unsigned long long)GPIO.OutputValue[1] << 7 |
		(unsigned long long)GPIO.OutputValue[2] << 8 |
		(unsigned long long)GPIO.OutputValue[3] << 9 |
		(unsigned long long)GPIO.OutputValue[4] << 10 |
		(unsigned long long)GPIO.OutputValue[5] << 11 |
		(unsigned long long)GPIO.DebounceEnable[0] << 12 |
		(unsigned long long)GPIO.DebounceEnable[1] << 13 |
		(unsigned long long)GPIO.DebounceEnable[2] << 14 |
		(unsigned long long)GPIO.DebounceEnable[3] << 15 |
		(unsigned long long)GPIO.DebounceEnable[4] << 16 |
		(unsigned long long)GPIO.DebounceEnable[5] << 17 |
		(unsigned long long)GPIO.Disconnect << 18 |
		((unsigned long long)GPIO.TriggerDebounceEnable & 0x1ull) << 19 |
		((unsigned long long)GPIO.TriggerDebounceEnable & 0x2ull) << 19);

	return(Status);
}

// Set the time of day (TOD) clock and load the TSI/TSF codes.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long LoadTODSettings(s_Adapter *p_HW, s_TODSettings TODClock)
{
	unsigned long Status = 0;

	Write(p_HW, BAR2, TOD_SETTINGS,
		(unsigned long long)TODClock.TODSeconds |
		(unsigned long long)TODClock.ClockFrequency << 32);

	Write(p_HW, BAR2, TSI_TSF_CODE,
		(unsigned long long)TODClock.TSFCode |
		(unsigned long long)TODClock.TSICode << 2);

	return(Status);
}

// Pause for the specified time in microseconds.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long MicrosecondTimer(unsigned long Time)
{
	// Pause for the specified time.
	return(Adapter_uSleep(Time));
}

// Report the model number of the device.
// Returns the model number in hex format, but it is not a hex value (e.g. Model 266 returns 0x266).
unsigned long ModelNumber(s_Adapter *p_HW)
{
	unsigned long long Value = 0;

	Read(p_HW, BAR2, IDENTIFICATION, &Value);

	return((unsigned long)((Value >> 32) & 0x0000000000000FFFull));
}

// Perform a 64-bit register read from the device at the specified BAR and address offset.
// Note: This function applies to both 32-bit and 64-bit operating systems.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long Read(s_Adapter *p_HW, unsigned long Bar, unsigned long Off, unsigned long long *p_Value)
{
	unsigned long Status = 0;

	// Read the requested offset and store the result in Value.
	Status = Adapter_Read64(p_HW, Bar, Off, p_Value);
	return(Status);
}

// Program a device through a serial interface port.
// Note: The result of a serial read operation can be accessed through the ReadValue pointer.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long Serial(s_Adapter *p_HW, unsigned long BusSelect, unsigned long MultiFunction, unsigned long Instruction, size_t InstructionSize, unsigned long Data, size_t DataSize, unsigned long *p_ReadValue)
{
	unsigned long Status = 0;
	unsigned long long config_payload = 0;
	unsigned long long max = 0xFFFFFFFFFFFFFFFFull;
	unsigned long long Value = 0;

	// Assemble the configuration settings.
	switch (InstructionSize+DataSize)
	{
		case 8:  {config_payload = ((unsigned long long)BusSelect << 4) | ((unsigned long long)MultiFunction << 3) | 0x0ull; break;}
		case 16: {config_payload = ((unsigned long long)BusSelect << 4) | ((unsigned long long)MultiFunction << 3) | 0x1ull; break;}
		case 24: {config_payload = ((unsigned long long)BusSelect << 4) | ((unsigned long long)MultiFunction << 3) | 0x2ull; break;}
		case 32: {config_payload = ((unsigned long long)BusSelect << 4) | ((unsigned long long)MultiFunction << 3) | 0x3ull; break;}
		case 40: {config_payload = ((unsigned long long)BusSelect << 4) | ((unsigned long long)MultiFunction << 3) | 0x4ull; break;}
		case 48: {config_payload = ((unsigned long long)BusSelect << 4) | ((unsigned long long)MultiFunction << 3) | 0x5ull; break;}
		default: {Status++; break;}
	}

	Status += Write(p_HW, BAR0, SERIAL_CONFIG, config_payload);

	// Load the payload bits.
	Status += Write(p_HW, BAR0, SERIAL_PAYLOAD, (unsigned long long)Instruction << DataSize | (unsigned long long)Data);

	// Initiate a serial transaction by writing any value to the serial command address.
	Status += Write(p_HW, BAR0, SERIAL_COMMAND, 0x0ull);

	// Wait for the serial transaction to complete by monitoring the busy bit.
	do
		Status += Read(p_HW, BAR0, SERIAL_COMMAND, &Value);
	while(Value);
	
	// Read the payload register to retrieve the results of a serial read operation.
	Status += Read(p_HW, BAR0, SERIAL_PAYLOAD, &Value);
	// Mask the bits to match the payload size for serial protocols other than I2C.
	*p_ReadValue = (unsigned long)Value & (max >> (64-DataSize));

	// Return an error code if the program operation did not complete.
	if (Status)
		return(1);
	else
		return(0);
}

// Issue a software reset that is equivalent to a power-on reset.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long SoftwareReset(s_Adapter *p_HW)
{
	return(Write(p_HW, BAR0, SOFTWARE_RESET, 0x0ull));
}

// Issue a software trigger that is synchronous to all channels.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long SoftwareTrigger(s_Adapter *p_HW)
{
	return(Write(p_HW, BAR2, SOFTWARE_TRIGGER, 0x0ull));
}

// Perform a 64-bit target PCI write.
// Note: This function applies to both 32-bit and 64-bit operating systems.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long Write(s_Adapter *p_HW, unsigned long Bar, unsigned long Off, unsigned long long Value)
{
	// Write the contents of Value in the specified offset.
	return(Adapter_Write64(p_HW, Bar, Off, Value));
}

