/****************************************************  Product Functions  *****************************************************

	Red Rapids
	797 North Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following functions are used to initialize the hardware specific to each SigFPGA model.

	Rev History:  11-11-2015:	Corrected Model 276 initialization.
				  01-16-2015:	Added support for Model 271.
				  10-20-2014:	Inital release.

	Copyright 2014, Red Rapids, Inc.

****************************************************  Product Functions  *****************************************************/

#include <stdlib.h>

#include "adapter.h"
#include "sigstream.h"
#include "dma_functions.h"
#include "common_functions.h"
#include "channel_functions.h"
#include "device_functions.h"
#include "product_functions.h"

// Intitialize the Model 271 hardware.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long M271Initialize(s_Adapter *p_HW, s_M271Devices M271Devices, s_Channel *p_RXChannel)
{
	unsigned long Status = 0;
	unsigned long chancnt = 0;
	size_t cnt = 0;
	s_User *User = NULL;

	// Cast the user defined pointer from the hardware handle.
	User = (s_User *)(p_HW->User);

	// The Model 377 can be equipped with either two or four receiver channels and no transmitter channels.
	User->RXChannelCount = ChannelCount(p_HW, 0);
	User->TXChannelCount = ChannelCount(p_HW, 1);
	// The Model 377 RX channels produce a 16-bit data item size.
	User->RXDataItemSize = 16;
	User->TXDataItemSize = 0;

	// Set the default values for Data Item Size, Item Packing Field Size, Resampler, and Input Select if they are set to zero in the structure.
	for (chancnt = 1; chancnt <= User->RXChannelCount; chancnt++)
	{
		if (!p_RXChannel[chancnt-1].PayloadFormat.DataItemSize) p_RXChannel[chancnt-1].PayloadFormat.DataItemSize = 16;
		if (!p_RXChannel[chancnt-1].PayloadFormat.ItemPackingFieldSize) p_RXChannel[chancnt-1].PayloadFormat.ItemPackingFieldSize = 16;
		if (!p_RXChannel[chancnt-1].DatapathControl.Transform[1]) p_RXChannel[chancnt-1].DatapathControl.Transform[1] = 1;
		if (!p_RXChannel[chancnt-1].SwitchSelect.ChannelSelect) p_RXChannel[chancnt-1].SwitchSelect.ChannelSelect = chancnt;
	}

	// Configure the I/O signal direction and impedance of the command/status port expander.
	Status += FXL6408Write(p_HW, CMD_STAT1, 0x3, 0xF8);
	Status += FXL6408Write(p_HW, CMD_STAT1, 0x7, 0x07);

	// Configure the first ADC that is assigned to RX channels #1 and #2 by loading the default values.
	for (cnt = 0; cnt < 21; cnt++)
		Status += ADS42LB69Write(p_HW, M271_RX1TO2_ADC, M271Devices.RX1TO2_ADC.Payload[cnt][0], M271Devices.RX1TO2_ADC.Payload[cnt][1]);

	// Configure the trim DACs from the default structure values.
	// Note: Trim DACs are only present on DC coupled units.
	for (cnt = 0; cnt < 3; cnt++)
	{
		Status += LTC1661Write(p_HW, M271_RX1_TRIM_DAC, M271Devices.RX1_TRIM_DAC.Payload[cnt][0], M271Devices.RX1_TRIM_DAC.Payload[cnt][1]);
		Status += LTC1661Write(p_HW, M271_RX2_TRIM_DAC, M271Devices.RX2_TRIM_DAC.Payload[cnt][0], M271Devices.RX2_TRIM_DAC.Payload[cnt][1]);
	}

	// Configure the amplifier enables from the default structure values.
	// Note: Amplifiers are only present on DC coupled units.
	for (cnt = 0; cnt < 2; cnt++)
		Status += ADL5566Write(p_HW, CMD_STAT1, M271Devices.RX1TO2_AMP_ENBL.Enable[cnt][0], M271Devices.RX1TO2_AMP_ENBL.Enable[cnt][1]);

	// Determine whether the external sample clock is active.
	if (ClockStatus(p_HW, CMD_STAT1) < 2)
		// An external clock was detected, so change the selected input to the clock distribution chip.
		// Note: The default input is the on-board synthesizer.
		M271Devices.ClockDist.Payload[9][1] = 0x02;

	// Hardware reset the clock distribution chip.
	Status += AD9512Write(p_HW, M271_CLK_DIST, 0x0000, 0x30);
	Status += AD9512Write(p_HW, M271_CLK_DIST, 0x005A, 0x01);
	Status += AD9512Write(p_HW, M271_CLK_DIST, 0x0000, 0x10);
	Status += AD9512Write(p_HW, M271_CLK_DIST, 0x005A, 0x01);
	// Configure the clock distribution chip by loading the default values.
	for (cnt = 0; cnt < 22; cnt++)
		Status += AD9512Write(p_HW, M271_CLK_DIST, M271Devices.ClockDist.Payload[cnt][0], M271Devices.ClockDist.Payload[cnt][1]);

	return(Status);
}

// Intitialize the Model 276 hardware.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long M276Initialize(s_Adapter *p_HW, s_M276Devices M276Devices, s_Channel *p_RXChannel)
{
	unsigned long Status = 0;
	unsigned long chancnt = 0;
	size_t cnt = 0;
	s_User *User = NULL;

	// Cast the user defined pointer from the hardware handle.
	User = (s_User *)(p_HW->User);

	// The standard Model 276 has two receiver channels and no transmitter channels.
	User->RXChannelCount = ChannelCount(p_HW, 0);
	User->TXChannelCount = ChannelCount(p_HW, 1);
	// The Model 276 RX channels produce a 16-bit data item size.
	User->RXDataItemSize = 12;
	User->TXDataItemSize = 0;

	// Set the default values for Data Item Size, Item Packing Field Size, Resampler, and Input Select if they are set to zero in the structure.
	for (chancnt = 1; chancnt <= User->RXChannelCount; chancnt++)
	{
		if (!p_RXChannel[chancnt-1].PayloadFormat.DataItemSize) p_RXChannel[chancnt-1].PayloadFormat.DataItemSize = 12;
		if (!p_RXChannel[chancnt-1].PayloadFormat.ItemPackingFieldSize) p_RXChannel[chancnt-1].PayloadFormat.ItemPackingFieldSize = 12;
		if (!p_RXChannel[chancnt-1].DatapathControl.Transform[1]) p_RXChannel[chancnt-1].DatapathControl.Transform[1] = 1;
		if (!p_RXChannel[chancnt-1].SwitchSelect.ChannelSelect) p_RXChannel[chancnt-1].SwitchSelect.ChannelSelect = chancnt;
	}

	// Configure the I/O signal direction and impedance of the command/status port expander.
	Status += FXL6408Write(p_HW, CMD_STAT1, 0x3, 0xF8);
	Status += FXL6408Write(p_HW, CMD_STAT1, 0x7, 0x07);

	// Configure the ADC from the default structure values.
	for (cnt = 0; cnt < 11; cnt++)
		Status += ADC12D1600Write(p_HW, M276_RX1TO2_ADC, M276Devices.RX1TO2_ADC.Payload[cnt][0], M276Devices.RX1TO2_ADC.Payload[cnt][1]);

	// Configure the trim DACs from the default structure values.
	// Note: Trim DACs are only present on DC coupled units.
	for (cnt = 0; cnt < 3; cnt++)
	{
		Status += LTC1661Write(p_HW, M276_RX1_TRIM_DAC, M276Devices.RX1_TRIM_DAC.Payload[cnt][0], M276Devices.RX1_TRIM_DAC.Payload[cnt][1]);
		Status += LTC1661Write(p_HW, M276_RX2_TRIM_DAC, M276Devices.RX2_TRIM_DAC.Payload[cnt][0], M276Devices.RX2_TRIM_DAC.Payload[cnt][1]);
	}

	// Configure the amplifier enables from the default structure values.
	// Note: Amplifiers are only present on DC coupled units.
	for (cnt = 0; cnt < 2; cnt++)
		Status += ADL5566Write(p_HW, CMD_STAT1, M276Devices.RX1TO2_AMP_ENBL.Enable[cnt][0], M276Devices.RX1TO2_AMP_ENBL.Enable[cnt][1]);

	return(Status);
}

// Intitialize the Model 277 hardware.
// Return of zero indicates success, non-zero indicates an error condition.
unsigned long M277Initialize(s_Adapter *p_HW, s_M277Devices M277Devices, s_Channel *p_RXChannel)
{
	unsigned long Status = 0;
	unsigned long chancnt = 0;
	size_t cnt = 0;
	s_User *User = NULL;

	// Cast the user defined pointer from the hardware handle.
	User = (s_User *)(p_HW->User);

	// The Model 377 can be equipped with either two or four receiver channels and no transmitter channels.
	User->RXChannelCount = ChannelCount(p_HW, 0);
	User->TXChannelCount = ChannelCount(p_HW, 1);
	// The Model 377 RX channels produce a 16-bit data item size.
	User->RXDataItemSize = 16;
	User->TXDataItemSize = 0;

	// Set the default values for Data Item Size, Item Packing Field Size, Resampler, and Input Select if they are set to zero in the structure.
	for (chancnt = 1; chancnt <= User->RXChannelCount; chancnt++)
	{
		if (!p_RXChannel[chancnt-1].PayloadFormat.DataItemSize) p_RXChannel[chancnt-1].PayloadFormat.DataItemSize = 16;
		if (!p_RXChannel[chancnt-1].PayloadFormat.ItemPackingFieldSize) p_RXChannel[chancnt-1].PayloadFormat.ItemPackingFieldSize = 16;
		if (!p_RXChannel[chancnt-1].DatapathControl.Transform[1]) p_RXChannel[chancnt-1].DatapathControl.Transform[1] = 1;
		if (!p_RXChannel[chancnt-1].SwitchSelect.ChannelSelect) p_RXChannel[chancnt-1].SwitchSelect.ChannelSelect = chancnt;
	}

	// Configure the I/O signal direction and impedance of the command/status port expander.
	Status += FXL6408Write(p_HW, CMD_STAT1, 0x3, 0xF8);
	Status += FXL6408Write(p_HW, CMD_STAT1, 0x7, 0x07);

	// Configure the first ADC that is assigned to RX channels #1 and #2 by loading the default values.
	for (cnt = 0; cnt < 21; cnt++)
		Status += ADS42LB69Write(p_HW, M277_RX1TO2_ADC, M277Devices.RX1TO2_ADC.Payload[cnt][0], M277Devices.RX1TO2_ADC.Payload[cnt][1]);

	// Configure the second ADC that is assigned to RX channels #3 and #4 by loading the default values.
	for (cnt = 0; cnt < 21; cnt++)
		Status += ADS42LB69Write(p_HW, M277_RX3TO4_ADC, M277Devices.RX3TO4_ADC.Payload[cnt][0], M277Devices.RX3TO4_ADC.Payload[cnt][1]);

	// Configure the trim DACs from the default structure values.
	// Note: Trim DACs are only present on DC coupled units.
	for (cnt = 0; cnt < 3; cnt++)
	{
		Status += LTC1661Write(p_HW, M277_RX1_TRIM_DAC, M277Devices.RX1_TRIM_DAC.Payload[cnt][0], M277Devices.RX1_TRIM_DAC.Payload[cnt][1]);
		Status += LTC1661Write(p_HW, M277_RX2_TRIM_DAC, M277Devices.RX2_TRIM_DAC.Payload[cnt][0], M277Devices.RX2_TRIM_DAC.Payload[cnt][1]);
		Status += LTC1661Write(p_HW, M277_RX3_TRIM_DAC, M277Devices.RX3_TRIM_DAC.Payload[cnt][0], M277Devices.RX3_TRIM_DAC.Payload[cnt][1]);
		Status += LTC1661Write(p_HW, M277_RX4_TRIM_DAC, M277Devices.RX4_TRIM_DAC.Payload[cnt][0], M277Devices.RX4_TRIM_DAC.Payload[cnt][1]);
	}

	// Configure the amplifier enables from the default structure values.
	// Note: Amplifiers are only present on DC coupled units.
	for (cnt = 0; cnt < 2; cnt++)
	{
		Status += ADL5566Write(p_HW, CMD_STAT1, M277Devices.RX1TO2_AMP_ENBL.Enable[cnt][0], M277Devices.RX1TO2_AMP_ENBL.Enable[cnt][1]);
		Status += ADL5566Write(p_HW, CMD_STAT1, M277Devices.RX3TO4_AMP_ENBL.Enable[cnt][0], M277Devices.RX3TO4_AMP_ENBL.Enable[cnt][1]);
	}

	// Determine whether the external sample clock is active.
	if (ClockStatus(p_HW, CMD_STAT1) < 2)
		// An external clock was detected, so change the selected input to the clock distribution chip.
		// Note: The default input is the on-board synthesizer.
		M277Devices.ClockDist.Payload[9][1] = 0x02;

	// Hardware reset the clock distribution chip.
	Status += AD9512Write(p_HW, M277_CLK_DIST, 0x0000, 0x30);
	Status += AD9512Write(p_HW, M277_CLK_DIST, 0x005A, 0x01);
	Status += AD9512Write(p_HW, M277_CLK_DIST, 0x0000, 0x10);
	Status += AD9512Write(p_HW, M277_CLK_DIST, 0x005A, 0x01);
	// Configure the clock distribution chip by loading the default values.
	for (cnt = 0; cnt < 22; cnt++)
		Status += AD9512Write(p_HW, M277_CLK_DIST, M277Devices.ClockDist.Payload[cnt][0], M277Devices.ClockDist.Payload[cnt][1]);

	return(Status);
}
